from __future__ import annotations

import functools
from inspect import isclass
from typing import Any, Callable, Dict, Type, TypeVar

from unipipe.utils.compat import get_annotations


def resolve_annotations(obj: Callable) -> Callable:
    """Returns an identical Callable object, but with all stringified type annotations
    resolved to Type objects.

    NOTE: KFP requires us to de-stringify type annotations before compiling
    components/pipelines.  It's also helpful for other backends/executors, since
    we no longer have to worry about de-stringify annotations each type we
    perform a type check.
    """
    setattr(obj, "__annotations__", get_annotations(obj, eval_str=True))
    return obj


def infer_type(obj: Any) -> Type:
    from unipipe.dsl import Component, LazyAttribute, Pipeline

    if isinstance(obj, Component):
        return get_annotations(obj.func, eval_str=True)["return"]
    elif isinstance(obj, LazyAttribute):
        parent_type = infer_type(obj.parent)
        return get_annotations(parent_type, eval_str=True).get(obj.key)
    elif isinstance(obj, Pipeline):
        return infer_type(obj.return_value)
    else:
        return type(obj)


def infer_input_types(inputs: Dict) -> Dict:
    return {k: infer_type(v) for k, v in inputs.items() if k != "return"}


_T = TypeVar("_T", str, float, int, bool, tuple)


def cast_output_type(output: Any, _type: Type[_T]) -> _T:
    if _type is None:
        assert output is None
        return output
    elif isclass(_type) and issubclass(_type, tuple):
        annotations = get_annotations(_type, eval_str=True)
        _kwargs = {
            k: cast_output_type(v, annotations[k])
            for k, v in zip(_type._fields, output)  # type: ignore
        }
        return _type(**_kwargs)  # type: ignore
    else:
        return _type(output)


def wrap_cast_output_type(func: Callable, _type: Type[_T]) -> Callable[..., _T]:
    @functools.wraps(func)
    def wrapped(*args, **kwargs) -> _T:
        result = func(*args, **kwargs)
        return cast_output_type(result, _type)

    return wrapped



from typing import NamedTuple

from unipipe import dsl
from unipipe.utils.annotations import infer_input_types, infer_type, resolve_annotations
from unipipe.utils.compat import get_annotations


def _stringified_annotations(a: "int") -> "str":
    return a * "nonsense"


def test_get_annotations():
    func = _stringified_annotations

    str_annotations = get_annotations(func)
    annotations = get_annotations(func, eval_str=True)
    assert all(isinstance(v, str) for v in str_annotations.values())
    assert all(eval(v) == annotations[k] for k, v in str_annotations.items())


def test_resolve_annotations():
    # NOTE: The 'copy.deepcopy' method does not create a deep copy of methods/classes,
    # so 'resolve_annotations' is forced to modify the existing function in-place.
    # Collect the stringified annotations before resolving.
    str_annotations = get_annotations(_stringified_annotations)
    assert all(isinstance(v, str) for v in str_annotations.values())

    func = resolve_annotations(_stringified_annotations)
    annotations = get_annotations(func)

    assert all(isinstance(v, type) for v in annotations.values())
    assert all(eval(v) == annotations[k] for k, v in str_annotations.items())


@dsl.component
def _split_name(name: str) -> NamedTuple("Output", first=str, last=str):  # type: ignore
    names = name.split(" ")
    return names[0], names[-1]


@dsl.component
def _hello(first_name: str, last_name: str) -> str:
    return f"Seven blessings, {first_name} of house {last_name}!"


def test_infer_type():
    name = "Tyrion Lannister"
    assert infer_type(name) == str

    split_name = _split_name(name=name)
    assert issubclass(infer_type(split_name), tuple)

    assert infer_type(split_name[0]) == str
    assert infer_type(split_name[1]) == str
    first, last = split_name
    assert infer_type(first) == str
    assert infer_type(last) == str

    greeting = _hello(first_name=first, last_name=last)
    assert infer_type(greeting) == str


def test_infer_input_types():
    split_name = _split_name(name="Tyrion Lannister")
    assert infer_input_types(split_name.inputs) == {"name": str}

    first, last = split_name
    greeting = _hello(first_name=first, last_name=last)
    assert infer_input_types(greeting.inputs) == {"first_name": str, "last_name": str}



import pytest

from unipipe.utils import ops


def _nonsensical_function(a: int) -> str:
    return a * "nonsense"


def test_dispatch_registry():
    registry = ops.dispatch

    # Test adding a new function to the dispatch registry
    assert "_nonsensical_function" not in registry.d
    ops.dispatch(_nonsensical_function)
    assert "_nonsensical_function" in registry.d

    # Test that several of the expected functions exist in the registry
    for key in ["equal", "add", "int_", "str_"]:
        assert key in registry.d


def test_multiple_dispatch_getitem():
    multiple_dispatch = ops.dispatch["equal"]

    # Test getting a few valid function signatures:
    #   - (str, str)
    inputs = {"a": "hello", "b": "hello"}
    assert multiple_dispatch[inputs](**inputs) is True
    inputs = {"a": "hello", "b": "world"}
    multiple_dispatch[inputs](**inputs) is False
    #   - (int, float)
    inputs = {"a": 1, "b": 1.0}
    multiple_dispatch[inputs](**inputs) is True
    inputs = {"a": 1, "b": 2.0}
    multiple_dispatch[inputs](**inputs) is False
    #   - (tuple, tuple)
    inputs = {"a": (1, 2, 3), "b": (1, 2, 3)}
    multiple_dispatch[inputs](**inputs) is True
    inputs = {"a": (1, 2, 3), "b": (1, 2)}
    multiple_dispatch[inputs](**inputs) is False

    # Test that invalid signatures raise a TypeError:
    #   - (str, int)
    with pytest.raises(TypeError):
        multiple_dispatch[{"a": "hello", "b": 1}]
    #   - (int, tuple)
    with pytest.raises(TypeError):
        multiple_dispatch[{"a": 1, "b": (1, 2, 3)}]
    #   - (int, tuple)
    with pytest.raises(TypeError):
        multiple_dispatch[{"a": 1, "b": (1, 2, 3), "c": "world"}]


def test_multiple_dispatch_call():
    multiple_dispatch = ops.dispatch["equal"]

    # Test calling a few valid function signatures:
    #   - (str, str)
    assert multiple_dispatch(a="hello", b="hello") is True
    assert multiple_dispatch(a="hello", b="world") is False
    #   - (int, float)
    assert multiple_dispatch(a=1, b=1.0) is True
    assert multiple_dispatch(a=1, b=2.0) is False
    #   - (tuple, tuple)
    assert multiple_dispatch(a=(1, 2, 3), b=(1, 2, 3)) is True
    assert multiple_dispatch(a=(1, 2, 3), b=(1, 2)) is False

    # Test that invalid signatures raise a TypeError:
    #   - (str, int)
    with pytest.raises(TypeError):
        multiple_dispatch(a="hello", b=1)
    #   - (int, tuple)
    with pytest.raises(TypeError):
        multiple_dispatch(a=1, b=(1, 2, 3))
    #   - (int, tuple, str)
    with pytest.raises(TypeError):
        multiple_dispatch(a=1, b=(1, 2, 3), c="world")



from typing import List, Optional, Sequence

import click
from pydantic import parse_raw_as

from unipipe import dsl
from unipipe.executor import EXECUTOR_IMPORTS
from unipipe.utils.scripts import run_script as unipipe_run_script

DEFAULT_SEQUENCE = ("None",)


@click.group()
def unipipe():
    pass


def validate_hardware(ctx, param, value) -> Optional[dsl.Hardware]:
    if value is None:
        return None
    else:
        return parse_raw_as(dsl.Hardware, value)


@unipipe.command(context_settings={"ignore_unknown_options": True})
@click.argument("path", nargs=1)
@click.argument("args", nargs=-1)
@click.option(
    "-e",
    "--executor",
    "executor",
    default="python",
    type=click.Choice(list(EXECUTOR_IMPORTS.keys())),
    help="Executor to use for launching the script. Default: 'python'",
)
@click.option(
    "-n",
    "--name",
    "name",
    default=None,
    type=str,
    help="Name of the resulting pipeline. Default: None",
)
@click.option(
    "-b",
    "--base-image",
    "base_image",
    default=None,
    type=str,
    help="Base Docker image for the script. Defaults to 'fkodom/unipipe:latest'.",
)
@click.option(
    "-p",
    "--package-to-install",
    "packages_to_install",
    multiple=True,
    type=str,
    help="Add a Python package to install for this job. Ex: 'torch==1.10.0'",
    default=DEFAULT_SEQUENCE,
)
@click.option(
    "-i",
    "--pip-index-url",
    "pip_index_urls",
    multiple=True,
    type=str,
    help="Add a custom PyPI URL for installing packages. Ex: 'https://pypi.org/simple'",
    default=DEFAULT_SEQUENCE,
)
@click.option(
    "-h",
    "--hardware",
    "hardware",
    default=None,
    type=click.UNPROCESSED,
    callback=validate_hardware,
    help=(
        "Override hardware to use for this job. Must be provided as a raw JSON string. "
        'Ex: \'{"cpus": 4, "memory": "4G"}\''
    ),
)
def run_script(
    path: str,
    args: Sequence[str],
    executor: str,
    name: Optional[str] = None,
    # Component args
    base_image: Optional[str] = None,
    packages_to_install: Optional[Sequence[str]] = None,
    pip_index_urls: Optional[Sequence[str]] = None,
    hardware: Optional[str] = None,
):
    if pip_index_urls == DEFAULT_SEQUENCE:
        pip_index_urls = None
    if packages_to_install == DEFAULT_SEQUENCE:
        packages_to_install = None

    unipipe_run_script(
        path=path,
        args=args,
        executor=executor,
        name=name,
        # Component args
        base_image=base_image,
        packages_to_install=packages_to_install,
        pip_index_urls=pip_index_urls,
        hardware=hardware,
    )



"""
Functions that solely exist for backwards-compatibility with older Python versions.
At the time of writing, 'unipipe' supports back to Python 3.7.
"""

from __future__ import annotations

import functools
import sys
import types


def removeprefix(string: str, prefix: str) -> str:
    if string.startswith(prefix):
        string = string[len(prefix) :]
    return string


def removesuffix(string: str, suffix: str) -> str:
    if string.endswith(suffix):
        string = string[: -len(suffix)]
    return string


def get_annotations(obj, *, globals=None, locals=None, eval_str=False):  # noqa: C901
    """Copy-pasta from the 'inspect' module in Python>=3.10.

    The '__future__.annotations' module works by "stringify-ing" type annotations,
    so they can be lazily evaluated at runtime, rather than at the time the
    function is defined.  With 'eval_str=True', this function will evaluate the
    type annotations for Callable objects, even when those annotations have been
    stringified.  (That's the intended usage within 'flo', but I avoided changing
    the default value 'eval_str=True' to be consistent with Python>=3.10.)

    ----------------

    Compute the annotations dict for an object.
    obj may be a callable, class, or module.
    Passing in an object of any other type raises TypeError.
    Returns a dict.  get_annotations() returns a new dict every time
    it's called; calling it twice on the same object will return two
    different but equivalent dicts.
    This function handles several details for you:
      * If eval_str is true, values of type str will
        be un-stringized using eval().  This is intended
        for use with stringized annotations
        ("from __future__ import annotations").
      * If obj doesn't have an annotations dict, returns an
        empty dict.  (Functions and methods always have an
        annotations dict; classes, modules, and other types of
        callables may not.)
      * Ignores inherited annotations on classes.  If a class
        doesn't have its own annotations dict, returns an empty dict.
      * All accesses to object members and dict values are done
        using getattr() and dict.get() for safety.
      * Always, always, always returns a freshly-created dict.
    eval_str controls whether or not values of type str are replaced
    with the result of calling eval() on those values:
      * If eval_str is true, eval() is called on values of type str.
      * If eval_str is false (the default), values of type str are unchanged.
    globals and locals are passed in to eval(); see the documentation
    for eval() for more information.  If either globals or locals is
    None, this function may replace that value with a context-specific
    default, contingent on type(obj):
      * If obj is a module, globals defaults to obj.__dict__.
      * If obj is a class, globals defaults to
        sys.modules[obj.__module__].__dict__ and locals
        defaults to the obj class namespace.
      * If obj is a callable, globals defaults to obj.__globals__,
        although if obj is a wrapped function (using
        functools.update_wrapper()) it is first unwrapped.
    """
    if isinstance(obj, type):
        # class
        obj_dict = getattr(obj, "__dict__", None)
        if obj_dict and hasattr(obj_dict, "get"):
            ann = obj_dict.get("__annotations__", None)
            if isinstance(ann, types.GetSetDescriptorType):
                ann = None
        else:
            ann = None

        obj_globals = None
        module_name = getattr(obj, "__module__", None)
        if module_name:
            module = sys.modules.get(module_name, None)
            if module:
                obj_globals = getattr(module, "__dict__", None)
        obj_locals = dict(vars(obj))
        unwrap = obj
    elif isinstance(obj, types.ModuleType):
        # module
        ann = getattr(obj, "__annotations__", None)
        obj_globals = getattr(obj, "__dict__")
        obj_locals = None
        unwrap = None
    elif callable(obj):
        # this includes types.Function, types.BuiltinFunctionType,
        # types.BuiltinMethodType, functools.partial, functools.singledispatch,
        # "class funclike" from Lib/test/test_inspect... on and on it goes.
        ann = getattr(obj, "__annotations__", None)
        obj_globals = getattr(obj, "__globals__", None)
        obj_locals = None
        unwrap = obj
    else:
        raise TypeError(f"{obj!r} is not a module, class, or callable.")

    if ann is None:
        return {}

    if not isinstance(ann, dict):
        raise ValueError(f"{obj!r}.__annotations__ is neither a dict nor None")

    if not ann:
        return {}

    if not eval_str:
        return dict(ann)

    if unwrap is not None:
        while True:
            if hasattr(unwrap, "__wrapped__"):
                unwrap = unwrap.__wrapped__
                continue
            if isinstance(unwrap, functools.partial):
                unwrap = unwrap.func
                continue
            break
        if hasattr(unwrap, "__globals__"):
            obj_globals = unwrap.__globals__

    if globals is None:
        globals = obj_globals
    if locals is None:
        locals = obj_locals

    return_value = {
        key: value if not isinstance(value, str) else eval(value, globals, locals)
        for key, value in ann.items()
    }
    return return_value



import os
import tempfile
from typing import Any, Dict, Generator, Optional

import pytest

from unipipe import dsl
from unipipe.utils.scripts import (
    component_from_script,
    function_from_script,
    get_component_kwargs_from_docstring,
    get_component_kwargs_from_script,
    get_docstring_from_script,
)

TEMPDIR = tempfile.TemporaryDirectory()
ORDINARY_DOCSTRING = """
This is a docstring...
"""
COMPONENT_DOCSTRING = """
@dsl.component(
    hardware=dsl.Hardware(cpus=2, memory='1G'),
    packages_to_install=['numpy'],
    pip_index_urls=['https://pypi.org/simple'],
)
"""
INVALID_DOCSTRING = """
@dsl.component(
    hardware=dsl.Hardware(cpus=2, memory='1G'),
"""
SCRIPT_NAME = "main"
SCRIPT_TEMPLATE = """
{docstring}

print('Hello, world!')
"""


@pytest.fixture(
    scope="session",
    params=[
        COMPONENT_DOCSTRING,
        INVALID_DOCSTRING,
        ORDINARY_DOCSTRING,
        COMPONENT_DOCSTRING + ORDINARY_DOCSTRING,
        None,
    ],
)
def docstring(request) -> Generator[Optional[str], None, None]:
    yield request.param


@pytest.fixture(scope="session")
def script(docstring: Optional[str]) -> Generator[str, None, None]:
    path = os.path.join(TEMPDIR.name, f"{SCRIPT_NAME}.py")
    code = SCRIPT_TEMPLATE.format(docstring=f"'''{docstring}'''" if docstring else "")
    with open(path, "w") as f:
        f.write(code)

    yield path


@pytest.fixture(scope="session")
def component_kwargs(
    docstring: Optional[str],
) -> Generator[Optional[Dict[str, Any]], None, None]:
    if docstring is not None and COMPONENT_DOCSTRING in docstring:
        yield dict(
            hardware=dsl.Hardware(cpus=2, memory="1G"),
            packages_to_install=["numpy"],
            pip_index_urls=["https://pypi.org/simple"],
        )
    else:
        yield None


def test_get_docstring_from_script(script: str, docstring: Optional[str]):
    result = get_docstring_from_script(script)
    if result is None:
        assert docstring is None
    else:
        assert isinstance(docstring, str)
        assert isinstance(result, str)
        assert docstring.strip() == result.strip()


def test_get_component_kwargs_from_docstring(
    docstring: Optional[str], component_kwargs: Optional[Dict[str, Any]]
):
    if docstring is None:
        pass
    elif docstring == INVALID_DOCSTRING:
        with pytest.raises(SyntaxError):
            _ = get_component_kwargs_from_docstring(docstring)
    else:
        assert component_kwargs == get_component_kwargs_from_docstring(docstring)


def test_get_component_kwargs_from_script(
    script: str, docstring: Optional[str], component_kwargs: Optional[Dict[str, Any]]
):
    if docstring == INVALID_DOCSTRING:
        pass
        with pytest.raises(SyntaxError):
            _ = get_component_kwargs_from_script(script)
    else:
        assert component_kwargs == get_component_kwargs_from_script(script)


def test_function_from_script(script: str):
    func = function_from_script(script)
    assert func.__name__ == SCRIPT_NAME


def test_component_from_script(script: str, docstring: Optional[str]):
    if docstring == INVALID_DOCSTRING:
        with pytest.raises(SyntaxError):
            component_from_script(script)
    else:
        component_fn = component_from_script(script)
        component = component_fn(args=[])
        assert component.func.__name__ == SCRIPT_NAME



from __future__ import annotations

import ast
import importlib
import os
import random
import sys
import tempfile
from contextlib import ExitStack
from itertools import dropwhile
from typing import Any, Callable, Dict, Optional, Sequence

import unipipe
from unipipe import dsl
from unipipe.utils.compat import removeprefix, removesuffix

EXIT_STACK = ExitStack()


def get_docstring_from_script(path: str) -> Optional[str]:
    with open(path, "r") as f:
        tree = ast.parse(f.read())
    return ast.get_docstring(tree, clean=True)


def parse_component_kwargs_from_string(decorator: str):
    decorator = removeprefix(decorator, "@")
    decorator = removeprefix(decorator, "dsl.")
    decorator = removeprefix(decorator, "component")
    decorator = decorator.format(**os.environ)
    return eval(f"dict{decorator}")


def get_component_kwargs_from_docstring(docstring: str) -> Optional[Dict[str, Any]]:
    lines = docstring.split("\n")
    lines = list(dropwhile(lambda x: not x.strip().startswith("@dsl.component"), lines))
    lines = [line for line in lines if not line.strip().startswith("#")]
    if not lines:
        return None

    exc: Optional[SyntaxError] = None
    # TODO: Make this cleaner and more robust.
    # I'm relatively new to Python's 'ast' module, but it seems like it would
    # be possible to extract the args more elegantly.  This is very brute-force.
    #
    # Iteratively add more lines to the string we're trying to parse.  If it
    # throws a SyntaxError, then add another line and try again.  Otherwise,
    # we must ahve successfully parsed the component kwargs.
    for num_lines in range(len(lines)):
        try:
            return parse_component_kwargs_from_string("".join(lines[: num_lines + 1]))
        except SyntaxError as e:
            exc = e

    assert exc is not None
    raise exc


def get_component_kwargs_from_script(script_path: str) -> Optional[Dict[str, Any]]:
    docstring = get_docstring_from_script(script_path)
    if docstring is None:
        return None

    return get_component_kwargs_from_docstring(docstring)


FUNCTION_NAME = "script_component"
FUNCTION_CODE = """
from typing import *
from uuid import uuid1

import unipipe
from unipipe import dsl
from unipipe.dsl import *


def {function_name}(args: List[str]) -> str:
    import argparse
    import sys

    sys.argv = ["{function_name}.py", *args]
    __name__ = "__main__"

    {script_code}

    return str(uuid1())
"""


def _random_python_file_name() -> str:
    characters = "abcdefghijklmnopqrstuvwxyz_"
    chosen = random.choices(characters, k=16)
    return "".join(chosen) + ".py"


def function_from_script(script_path: str) -> Callable:
    name = removesuffix(os.path.basename(script_path), ".py")
    with open(script_path, "r") as f:
        script_lines = f.readlines()
    indent = " " * 4
    code = FUNCTION_CODE.format(
        function_name=name, script_code=indent.join(script_lines)
    )

    tempdir = EXIT_STACK.enter_context(tempfile.TemporaryDirectory())
    file_name = _random_python_file_name()
    path = os.path.join(tempdir, file_name)
    with open(path, "w") as f:
        f.write(code)
        f.flush()

    if tempdir not in sys.path:
        sys.path.append(tempdir)
    module = importlib.import_module(removesuffix(file_name, ".py"))

    return getattr(module, name)


def component_from_script(path: str, **manual_kwargs) -> Callable[..., dsl.Component]:
    # NOTE: Get the default kwargs first!  If there is a syntax error in the docstring,
    # we prefer this function to throw a 'SyntaxError'.
    default_kwargs = get_component_kwargs_from_script(path)
    if default_kwargs is None:
        default_kwargs = {}

    func = function_from_script(path)

    # Remove any 'None' or empty values that may have been included by default
    # or forwarded from the CLI tool.
    default_kwargs = {k: v for k, v in default_kwargs.items() if v is not None}
    manual_kwargs = {k: v for k, v in manual_kwargs.items() if v is not None}
    kwargs = {**default_kwargs, **manual_kwargs}

    return dsl.component(func=func, **kwargs)


def run_script(
    path: str,
    args: Sequence[str] = (),
    executor: str = "python",
    name: Optional[str] = None,
    **kwargs,
):
    component_fn = component_from_script(path=path, **kwargs)
    # Component functions expect a 'List' object. It's easier to type-check a strict
    # 'List' annotation than 'Sequence', but we want the 'run_script' method to be as
    # user-friendly as possible.
    args = list(args)

    @dsl.pipeline(name=name)
    def pipeline():
        component_fn(args)

    unipipe.run(executor=executor, pipeline=pipeline())



from importlib import import_module
from typing import Dict, Optional, Union

from pydantic import BaseModel

from unipipe.dsl import Pipeline
from unipipe.executor.base import Executor


class ExecutorImport(BaseModel):
    module: str
    name: str


EXECUTOR_IMPORTS: Dict[str, ExecutorImport] = {
    "docker": ExecutorImport(module="unipipe.executor.docker", name="DockerExecutor"),
    "python": ExecutorImport(module="unipipe.executor.python", name="PythonExecutor"),
    "vertex": ExecutorImport(module="unipipe.executor.vertex", name="VertexExecutor"),
}


def run(
    executor: Union[str, Executor],
    pipeline: Pipeline,
    **kwargs,
):
    if isinstance(executor, str):
        _import = EXECUTOR_IMPORTS[executor]
        executor = getattr(import_module(_import.module), _import.name)()
        assert isinstance(executor, Executor)
    return executor.run(pipeline, **kwargs)



import json
import logging
import os
import sys
import tempfile
import textwrap
from inspect import getsource, isclass
from itertools import dropwhile
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Type, Union

from docker.errors import BuildError
from docker.types import DeviceRequest

import docker
from unipipe.dsl import Component, ConditionalPipeline, LazyAttribute, Pipeline
from unipipe.executor.base import LocalExecutor
from unipipe.utils.compat import get_annotations

if sys.version_info >= (3, 8):
    from typing import TypedDict  # pylint: disable=no-name-in-module
else:
    from typing_extensions import TypedDict


IMPORTS = """
import typing
from typing import *

import unipipe
from unipipe import dsl
from unipipe.dsl import *
from unipipe.utils.ops import dispatch
"""

LOGGING = """
import logging
logging.getLogger().setLevel({logging_level})
"""

ARGPARSE_LIST = "argparse_list"
COMMAND = """
import argparse
import ast
import json

def {argparse_list}(s):
    data = ast.literal_eval(s)
    print(s, data)
    return tuple(str(x) for x in data)

parser = argparse.ArgumentParser()
{arguments}
args = parser.parse_args()

output = {function_name}(**vars(args))
if isinstance(output, dsl.Component):
    output = output.func(**vars(args))

with open('/app/output.json', "w") as f:
    json.dump(dict(output=output), f)
"""


def _get_component_func_source(func: Callable) -> str:
    """Largely copy-pasta from 'kfp.v2'.  De-indents the function source code, and
    removes any decorators or other code preceding the 'def' statement.  Then,
    decorate the function with just '@dsl.component', so we can utilize type
    checking/casting, logging, etc. from the Component class.
    """
    # Function may be defined in another function/class. Dedent the source code.
    lines = textwrap.dedent(getsource(func)).split("\n")
    lines = list(dropwhile(lambda x: not x.startswith("def"), lines))

    if not lines:
        raise ValueError(
            'Failed to dedent and clean up the source of function "{}". '
            "It is probably not properly indented.".format(func.__name__)
        )

    return "\n".join(["@dsl.component", *lines])


def _get_argparse_argument(name: str, annotation: Type) -> str:
    if hasattr(annotation, "__name__"):
        if annotation in (List[str], Tuple[str]):
            # As of Python 3.10, the List/Tuple classes have a '__name__' property
            args = f"type={ARGPARSE_LIST}"
        else:
            args = f"type={annotation.__name__}"
    elif hasattr(annotation, "_name"):
        args = f"type={ARGPARSE_LIST}"
    else:
        raise TypeError(f"Could not determine a valid type string for '{annotation}'.")

    return f"parser.add_argument('--{name}', {args})"


def build_script(component: Component) -> str:
    _logging = LOGGING.format(logging_level=component.logging_level)
    function = _get_component_func_source(component.func)
    annotations = get_annotations(component.func, eval_str=True)
    argument_lines = [
        _get_argparse_argument(name=k, annotation=v)
        for k, v in annotations.items()
        if k != "return"
    ]
    command = COMMAND.format(
        argparse_list=ARGPARSE_LIST,
        arguments="\n".join(argument_lines),
        function_name=component.func.__name__,
    )
    return "\n".join([IMPORTS, _logging, function, command])


# TODO: Once stable release is available in PyPI:
#   +++ pip install unipipe
#   --- pip install ./unipipe

DOCKERFILE = """
FROM {base_image}

WORKDIR /app
COPY ./ ./unipipe
RUN pip install ./unipipe
RUN pip install {packages}
"""


def _record_build_logs(logs: Iterable[Dict[str, str]], level: int):
    for line in logs:
        if "stream" in line:
            logging.log(level=level, msg=line["stream"])


def build_docker_image(component: Component, tag: str):
    base_image = component.base_image
    logging.info(f"Building Docker image: ('tag={tag}', 'base_image={base_image}')")
    client = docker.from_env()

    # If no packages listed, just include 'pip' as a placeholder
    install_options = component.packages_to_install or ["pip"]
    if component.pip_index_urls:
        index_url, *extra_index_urls = component.pip_index_urls
        install_options.append(f"--index-url {index_url} --trusted-host {index_url}")
        install_options.extend(
            [f"--extra-index-url {i} --trusted-host {i}" for i in extra_index_urls]
        )

    packages = " ".join(install_options)
    dockerfile = DOCKERFILE.format(base_image=base_image, packages=packages)

    with tempfile.TemporaryDirectory() as tempdir:
        dockerfile_path = os.path.join(tempdir, "Dockerfile")
        with open(dockerfile_path, "w") as f:
            f.write(dockerfile)
        try:
            _, logs = client.images.build(
                path="./", dockerfile=dockerfile_path, tag=tag, pull=True, rm=True
            )
            level = logging.DEBUG
        except BuildError as e:
            logs = e.build_log
            level = logging.ERROR

    _record_build_logs(logs, level=level)
    return tag


class Volume(TypedDict):
    bind: str
    mode: str


def _get_cli_argument(name: str, value: Any) -> str:
    if isinstance(value, (list, tuple)):
        arg_strings = [f"'{v}'" for v in value]
        arg_list_string = "[" + ", ".join(arg_strings) + "]"
        return f'--{name}="{arg_list_string}"'
    else:
        return f"--{name}='{value}'"


def build_and_run(
    component: Component,
    arguments: Optional[Dict[str, Any]] = None,
    volumes: Optional[Dict[str, Union[Dict, Volume]]] = None,
    remove: bool = True,
):
    client = docker.from_env()
    tag = build_docker_image(component, tag=component.name)
    if arguments is None:
        arguments = {}

    if volumes is None:
        volumes = {}

    config_path = os.path.expanduser("~/.config/")
    if os.path.exists(config_path) and config_path not in volumes:
        volumes[config_path] = {"bind": "/root/.config/", "mode": "ro"}

    with tempfile.TemporaryDirectory() as tempdir:
        script_path = os.path.join(tempdir, "main.py")
        output_json = os.path.join(tempdir, "output.json")
        script = build_script(component)
        with open(script_path, "w") as f:
            f.write(script)

        volumes[tempdir] = {"bind": "/app/", "mode": "rw"}
        args = " ".join(
            [_get_cli_argument(name=k, value=v) for k, v in arguments.items()]
        )

        device_requests = []
        accelerator = component.hardware.accelerator
        if accelerator is not None and accelerator.count:
            # TODO:
            #   - Make this logic work for TPUs as well
            #   - Allow users to pick specific device IDs?
            device_ids = list(range(int(accelerator.count)))
            device_requests.append(
                DeviceRequest(
                    device_ids=[",".join([str(i) for i in device_ids])],
                    capabilities=[["gpu"]],
                )
            )

        container = client.containers.run(
            image=component.name,
            command=f"python /app/main.py {args}",
            volumes=volumes,
            remove=remove,
            detach=True,
            device_requests=device_requests,
        )
        for line in container.logs(stream=True):
            line = line.strip()
            if line:
                print(line.decode("utf-8"))

        container.wait()
        client.images.remove(tag, force=True, noprune=False)

        with open(output_json, "r") as f:
            result = json.load(f)["output"]

    return result


def resolve_value(arguments: Dict, value: Any) -> Any:
    if isinstance(value, LazyAttribute):
        return getattr(resolve_value(arguments, value.parent), value.key)
    elif isinstance(value, Pipeline):
        if value.name in arguments:
            return arguments[value.name]
        return resolve_value(arguments, value.return_value)
    elif isinstance(value, (tuple, list)):
        return tuple(resolve_value(arguments, x) for x in value)
    elif isinstance(value, Component):
        assert isinstance(arguments, dict)
        return arguments[value.name]
    else:
        return value


class DockerExecutor(LocalExecutor):
    def resolve_local_value(self, _locals: Dict, value: Any) -> Any:
        if isinstance(value, LazyAttribute):
            return getattr(resolve_value(_locals, value.parent), value.key)
        elif isinstance(value, Pipeline):
            if value.name in _locals:
                return _locals[value.name]
            return resolve_value(_locals, value.return_value)
        elif isinstance(value, (tuple, list)):
            return tuple(resolve_value(_locals, x) for x in value)
        elif isinstance(value, Component):
            assert isinstance(_locals, dict)
            return _locals[value.name]
        else:
            return value

    def run_component(self, component: Component, **kwargs):
        result = build_and_run(component, kwargs)
        return_type = get_annotations(component.func, eval_str=True).get("return")
        if isclass(return_type) and issubclass(return_type, tuple):
            result = return_type(*result)

        return result

    def run_conditional_pipeline_with_locals(
        self, pipeline: ConditionalPipeline, _locals: Dict[str, Any]
    ):
        operand1 = self.resolve_local_value(_locals, pipeline.condition.operand1)
        operand2 = self.resolve_local_value(_locals, pipeline.condition.operand2)
        comparator = pipeline.condition.comparator

        if comparator(operand1, operand2):
            return self.run_pipeline_with_locals(pipeline, _locals=_locals)
        else:
            return None, _locals



from __future__ import annotations

import logging
from contextlib import ExitStack, contextmanager
from enum import Enum
from functools import partial, wraps
from inspect import signature
from types import TracebackType
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
)
from uuid import uuid1

from pydantic import BaseModel, parse_obj_as

from unipipe.utils import ops
from unipipe.utils.annotations import wrap_cast_output_type
from unipipe.utils.compat import get_annotations

T_co = TypeVar("T_co", covariant=True)


class AcceleratorType(str, Enum):
    T4 = "nvidia-tesla-t4"
    V100 = "nvidia-tesla-v100"
    P4 = "nvidia-tesla-p4"
    P100 = "nvidia-tesla-p100"
    K80 = "nvidia-tesla-k80"


class Accelerator(BaseModel):
    count: Optional[int] = None
    type: Optional[AcceleratorType] = None

    class Config:
        allow_population_by_field_name: bool = True


class Hardware(BaseModel):
    """
    Hardware resources for a pipeline component.
    """

    cpus: Optional[str] = None
    memory: Optional[str] = None
    accelerator: Optional[Accelerator] = None

    class Config:
        allow_population_by_field_name: bool = True


# TODO: Open an issue with google about the minimum allowable hardware. Even though
# custom jobs allow the 'e2-standard-1' machine type, apparently Vertex Pipelines
# will only allocate 'e2-highmem-2' as the smallest machine type.
#
# Not a huge deal, but will cost you a few cents here and there for small jobs.
MINIMAL_HARDWARE = Hardware(cpus=1, memory="512M")


class _Operable:
    def __len__(self) -> Component:
        return dispatch_to_component(ops.len_, a=self)

    def __str__(self) -> Component:  # type: ignore
        return dispatch_to_component(ops.str_, a=self)

    def __int__(self) -> Component:  # type: ignore
        return dispatch_to_component(ops.int_, a=self)

    def __float__(self) -> Component:  # type: ignore
        return dispatch_to_component(ops.float_, a=self)

    def __add__(self, other: Any) -> Component:
        return dispatch_to_component(ops.add, a=self, b=other)

    def __sub__(self, other: Any) -> Component:
        return dispatch_to_component(ops.sub, a=self, b=other)

    def __mul__(self, other: Any) -> Component:
        return dispatch_to_component(ops.mul, a=self, b=other)

    def __div__(self, other: Any) -> Component:
        return dispatch_to_component(ops.div, a=self, b=other)

    def __eq__(self, other: Any) -> Component:  # type: ignore
        return dispatch_to_component(ops.equal, a=self, b=other)


class LazyAttribute(BaseModel, _Operable):  # type: ignore
    parent: Any
    key: str


class LazyItem(BaseModel, _Operable):  # type: ignore
    parent: Any
    idx: int


def wrap_logging_info(
    func: Callable, component_name: str, logging_level: int
) -> Callable:
    @wraps(func)
    def wrapped(*args, **kwargs):
        import logging

        logging.getLogger().setLevel(logging_level)
        result = func(*args, **kwargs)
        logging.info(f"[{component_name}] - {result}")
        return result

    return wrapped


def _base_image_for_hardware(hardware: Hardware) -> str:
    accelerator = hardware.accelerator
    if accelerator is not None and accelerator.count:
        return "fkodom/unipipe:latest-cuda"
    else:
        return "fkodom/unipipe:latest"


def get_pip_index_urls(urls: Optional[Sequence[str]]) -> Optional[List[str]]:
    if urls is None:
        return None

    PYPI_URL = "https://pypi.org/simple"
    result = [url for url in urls]
    if PYPI_URL not in result:
        result.append(PYPI_URL)

    return result


class Component(_Operable, Generic[T_co]):
    def __init__(
        self,
        func: Callable,
        name: Optional[str] = None,
        inputs: Optional[Dict] = None,
        logging_level: Optional[int] = None,
        # Not used by the local executor
        base_image: Optional[str] = None,
        packages_to_install: Optional[List[str]] = None,
        pip_index_urls: Optional[List[str]] = None,
        hardware: Optional[Union[Dict, Hardware]] = None,
    ) -> None:
        """
        Args:
            func: (Callable) Function that defines this component.
            name: (str) The name for this component.
            inputs: Optional(Dict) A dictionary containing the input values assigned to their parameter names.
        """
        if name is None:
            uuid = str(uuid1())[:8]
            name = f"{func.__name__}-{uuid}"
        inputs = inputs or {}

        self.name = name.replace("_", "-")
        logging_level = logging_level or logging.INFO
        self.return_type = get_annotations(func, eval_str=True)["return"]
        self.func = wrap_logging_info(
            wrap_cast_output_type(func, _type=self.return_type),
            component_name=name,
            logging_level=logging_level,
        )
        self.inputs = inputs
        self.logging_level = logging_level
        self.packages_to_install = packages_to_install
        self.pip_index_urls = get_pip_index_urls(pip_index_urls)
        self.hardware = parse_obj_as(Hardware, hardware) if hardware else Hardware()
        self.base_image = base_image or _base_image_for_hardware(self.hardware)

        pipeline = PipelineContext().current
        if pipeline is not None:
            pipeline.components.append(self)

    def _len(self) -> int:
        if issubclass(self.return_type, tuple):
            return len(self.return_type._fields)

        raise TypeError(
            "Only components with 'NamedTuple' return type have a defined length. "
            f"Found return type {self.return_type}."
        )

    def __iter__(self):
        return (self[i] for i in range(self._len()))

    def __getattr__(self, key: str) -> LazyAttribute:
        return LazyAttribute(parent=self, key=key)

    def __getitem__(self, idx: int) -> LazyAttribute:
        if issubclass(self.return_type, tuple):
            key = self.return_type._fields[idx]
            return LazyAttribute(parent=self, key=key)

        raise TypeError(
            "Only components with 'NamedTuple' return type have '__getitem__' method. "
            f"Found return type {self.return_type}."
        )


def _process_component_args(func: Callable, *args, **kwargs):
    sig = signature(func)
    pos_inputs = {k: a for k, a in zip(sig.parameters.keys(), args)}
    for key in pos_inputs.keys():
        if key in kwargs.keys():
            raise ValueError(
                f"Function '{func.__name__}' received two values for arg '{key}'."
            )

    component_inputs = {**pos_inputs, **kwargs}
    return component_inputs


def component(
    func: Optional[Callable] = None,
    name: Optional[str] = None,
    logging_level: Optional[int] = None,
    base_image: Optional[str] = None,
    packages_to_install: Optional[List[str]] = None,
    pip_index_urls: Optional[List[str]] = None,
    hardware: Optional[Union[Dict, Hardware]] = None,
) -> Callable:
    new_component = partial(
        Component,
        name=name,
        logging_level=logging_level,
        base_image=base_image,
        packages_to_install=packages_to_install,
        pip_index_urls=pip_index_urls,
        hardware=hardware,
    )

    if func is None:

        def wrapper(func: Callable) -> Callable[..., Component]:
            @wraps(func)
            def wrapped_component(*args, **kwargs) -> Component:
                inputs = _process_component_args(func, *args, **kwargs)
                return new_component(func=func, inputs=inputs)

            return wrapped_component

        return wrapper
    else:

        @wraps(func)
        def wrapped_component(*args, **kwargs) -> Component:
            assert func is not None
            inputs = _process_component_args(func, *args, **kwargs)
            return new_component(func=func, inputs=inputs)

        return wrapped_component


def dispatch_to_component(dispatch: ops.MultipleDispatch, **kwargs) -> Component:
    func = dispatch[kwargs]
    component_func = component(func=func, hardware=MINIMAL_HARDWARE)
    return component_func(**kwargs)


class Pipeline(ExitStack, _Operable, Generic[T_co]):
    def __init__(
        self,
        name: Optional[str] = None,
        components: Optional[List[Union[Component, Pipeline]]] = None,
        inputs: Optional[Dict] = None,
        return_value: Optional[Any] = None,
    ) -> None:
        """
        Args:
            name: Name of the pipeline.
            components: List of components to register to the pipeline.
        """
        super().__init__()
        if name is None:
            uuid = str(uuid1())[:8]
            name = f"{self.__class__.__name__.lower()}-{uuid}"

        self.name = name.replace("_", "-")  # Cannot use _ in pipeline names
        self.components = components or []
        self.inputs = inputs or {}
        self.return_value = return_value
        self.parent: Optional[Pipeline] = None

        context = PipelineContext()
        if context.current is not None:
            context.current.components.append(self)

    def __enter__(self):
        context = PipelineContext()
        self.parent = context.current
        context.current = self
        return self

    def __exit__(
        self,
        type: Optional[type[BaseException]],
        exc: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> bool:
        context = PipelineContext()
        context.current = self.parent
        return super().__exit__(type, exc, traceback)

    @property
    def return_type(self) -> Type:
        return type(self.return_value)

    def _len(self) -> int:
        if issubclass(self.return_type, (tuple, list, dict)):
            assert self.return_value is not None
            return len(self.return_value)

        raise TypeError(
            f"Length of 'return_value={self.return_value}' is not defined at runtime "
            "(not determined until execution time). Pipelines with return types "
            "(tuple, list, dict) are defined at runtime, though, and can be "
            "indexed/iterated like normal Python container objects."
        )

    def __iter__(self):
        return (self[i] for i in range(self._len()))

    def __getitem__(self, idx: int) -> LazyItem:
        return LazyItem(parent=self.return_value, idx=idx)

    def __getattr__(self, key: str) -> LazyAttribute:
        return LazyAttribute(parent=self, key=key)


class PipelineContext:
    current: Optional[Pipeline] = None

    def __new__(cls):
        if not hasattr(cls, "instance"):
            cls.instance = super().__new__(cls)
        return cls.instance


def pipeline(
    func: Optional[Callable] = None,
    name: Optional[str] = None,
    **kwargs,
):
    if func is None:

        def wrapper(func: Callable) -> Callable:
            def wrapped_pipeline(**inputs) -> Pipeline:
                with Pipeline(name=name, **kwargs) as pipe:
                    return_value = func(**inputs)
                    pipe.return_value = return_value
                return pipe

            return wrapped_pipeline

        return wrapper
    else:

        def wrapped_pipeline(**inputs) -> Pipeline:
            assert func is not None
            with Pipeline(name=name, **kwargs) as pipe:
                return_value = func(**inputs)
                pipe.return_value = return_value
            return pipe

        return wrapped_pipeline


class Condition(BaseModel):
    operand1: Any
    operand2: Any
    comparator: Callable


class ConditionalPipeline(Pipeline):
    def __init__(
        self,
        condition: Condition,
        name: Optional[str] = None,
        components: Optional[List[Union[Component, Pipeline]]] = None,
        inputs: Optional[Dict] = None,
        return_value: Optional[Any] = None,
    ) -> None:
        super().__init__(name, components, inputs, return_value)
        self.condition = condition


@contextmanager
def condition(
    operand1: Any,
    operand2: Any,
    comparator: Callable[[Any, Any], bool],
    name: Optional[str] = None,
):
    _condition = Condition(operand1=operand1, operand2=operand2, comparator=comparator)
    pipeline = ConditionalPipeline(name=name, condition=_condition)
    try:
        with pipeline:
            yield pipeline
    finally:
        pass


@wraps(condition)
def equal(operand1: Any, operand2: Any, name: Optional[str] = None):
    _uuid = uuid1()
    if name is None:
        name = f"equal_{_uuid}"
    return condition(operand1, operand2, comparator=lambda o1, o2: o1 == o2, name=name)


@wraps(condition)
def not_equal(operand1: Any, operand2: Any, name: Optional[str] = None):
    _uuid = uuid1()
    if name is None:
        name = f"equal_{_uuid}"
    return condition(operand1, operand2, comparator=lambda o1, o2: o1 != o2, name=name)


@wraps(condition)
def _depends_on(operand1: Any, name: Optional[str] = None):
    _uuid = uuid1()
    if name is None:
        name = f"depends_on_{_uuid}"
    return condition(
        operand1,
        operand2=str(_uuid),
        comparator=lambda o1, o2: o1 != o2,
        name=name,
    )


@contextmanager
def depends_on(*components: Component):
    stack = ExitStack()
    component_stack = stack.enter_context(ExitStack())

    def depends(*args):
        nonlocal component_stack
        condition = _depends_on(args[0])
        component_stack = component_stack.enter_context(condition)

        if len(args) > 1:
            return depends(*args[1:])

    depends(*components)
    yield stack



from __future__ import annotations

from abc import abstractmethod
from typing import Any, Dict, Tuple

from unipipe.dsl import Component, ConditionalPipeline, Pipeline


class Executor:
    @abstractmethod
    def run(self, pipeline: Pipeline):
        pass


class LocalExecutor(Executor):
    @abstractmethod
    def resolve_local_value(self, _locals: Dict, value: Any) -> Any:
        pass

    @abstractmethod
    def run_component(self, component: Component, **kwargs):
        pass

    @abstractmethod
    def run_conditional_pipeline_with_locals(
        self, pipeline: ConditionalPipeline, _locals: Dict[str, Any]
    ):
        pass

    def run_pipeline_with_locals(
        self, pipeline: Pipeline, _locals: Dict[str, Any]
    ) -> Tuple[Any, Dict[str, Any]]:
        for component in pipeline.components:
            kwargs = {
                k: self.resolve_local_value(_locals, v)
                for k, v in component.inputs.items()
            }
            __locals = {**_locals, **kwargs}
            if isinstance(component, ConditionalPipeline):
                result, _ = self.run_conditional_pipeline_with_locals(
                    component, _locals=__locals
                )
            elif isinstance(component, Pipeline):
                result, _ = self.run_pipeline_with_locals(component, _locals=__locals)
            elif isinstance(component, Component):
                result = self.run_component(component, **kwargs)
            else:
                raise TypeError(
                    f"Found pipeline component {component} with unexpected type: "
                    f"{type(component)}. Valid component types are "
                    "[Component, ConditionalPipeline, Pipeline]."
                )

            _locals[component.name] = result

        return_value = self.resolve_local_value(_locals, pipeline.return_value)
        return return_value, _locals

    def run(self, pipeline: Pipeline):
        return_value, _ = self.run_pipeline_with_locals(
            pipeline, _locals=pipeline.inputs
        )
        return return_value



import os
from distutils.core import setup
from subprocess import getoutput

import setuptools


def get_version_tag() -> str:
    try:
        env_key = "UNIPIPE_VERSION".upper()
        version = os.environ[env_key]
    except KeyError:
        version = getoutput("git describe --tags --abbrev=0")

    if version.lower().startswith("fatal"):
        version = "0.0.0"

    return version


extras_require = {
    "docker": ["docker"],
    "vertex": ["kfp", "google-cloud-aiplatform", "google-cloud-storage"],
    "test": [
        "arrow",
        "black",
        "flake8",
        "isort",
        "mypy",
        "pytest",
        "pytest-cov",
        "torch",
        "torchvision",
    ],
}
extras_require["dev"] = ["pre-commit", "ipywidgets", *extras_require["test"]]
all_require = [r for reqs in extras_require.values() for r in reqs]
extras_require["all"] = all_require


setup(
    name="unipipe",
    version=get_version_tag(),
    author="Frank Odom",
    author_email="frank.odom.iii@gmail.com",
    url="https://github.com/fkodom/unipipe",
    packages=setuptools.find_packages(exclude=["tests"]),
    description="project_description",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    install_requires=["click", "pydantic", "typing_extensions"],
    extras_require=extras_require,
    entry_points={
        "console_scripts": [
            "unipipe = unipipe.cli:unipipe",
        ],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.7",
)



import json
import os
import tempfile
from unittest import mock

import pytest
from kfp.v2.compiler import Compiler

from examples.ex01_hello_world import pipeline as pipeline_01
from examples.ex02_hello_pipeline import pipeline as pipeline_02
from examples.ex03_multi_output_components import pipeline as pipeline_03
from examples.ex04_pipeline_arguments import pipeline as pipeline_04
from examples.ex05_dependency_management import pipeline as pipeline_05
from examples.ex06_hardware_specs import pipeline as pipeline_06
from examples.ex07_nested_pipelines import pipeline as pipeline_07
from examples.ex08_control_flow import pipeline as pipeline_08
from examples.ex09_advanced_control_flow import good_pipeline as pipeline_09
from unipipe import dsl
from unipipe.backend.kfp import KubeflowPipelinesBackend
from unipipe.utils.scripts import component_from_script


def _test_build_kfp_pipeline(pipeline: dsl.Pipeline):
    kfp_pipeline = KubeflowPipelinesBackend().build(pipeline=pipeline)
    with tempfile.TemporaryDirectory() as tempdir:
        path = os.path.join(tempdir, "pipeline.json")
        Compiler().compile(pipeline_func=kfp_pipeline, package_path=path)

        # Check that the JSON file was created, and we can successfully load it.
        assert os.path.exists(path)
        with open(path) as f:
            _ = json.load(f)


def test_example_01():
    _test_build_kfp_pipeline(pipeline_01())


def test_example_02():
    _test_build_kfp_pipeline(pipeline_02())


def test_example_03():
    _test_build_kfp_pipeline(pipeline_03())


def test_example_04():
    _test_build_kfp_pipeline(pipeline_04(name="Tyrion Lannister"))


def test_example_05():
    _test_build_kfp_pipeline(pipeline_05(name="Tyrion Lannister"))


def test_example_06():
    image_url = "https://raw.githubusercontent.com/EliSchwartz/imagenet-sample-images/master/n01443537_goldfish.JPEG"
    _test_build_kfp_pipeline(pipeline_06(image_url=image_url))


def test_example_07():
    _test_build_kfp_pipeline(pipeline_07())


def test_example_08():
    _test_build_kfp_pipeline(pipeline_08())


def test_example_09():
    _test_build_kfp_pipeline(pipeline_09(name="Ned Stark"))


def test_example_11():
    # This will fail, because it doesn't have any PyPI credentials as ENV variables.
    # It should raise 'KeyError' as it tries to fetch those from 'os.environ'.
    with pytest.raises(KeyError):
        component_fn = component_from_script(path="./examples/ex11_using_scripts.py")

    # Set mock values, so we can run the script without affecting other tests.
    mock_pypi_credentials = {"PYPI_USERNAME": "user", "PYPI_PASSWORD": "pass"}
    with mock.patch.dict(os.environ, mock_pypi_credentials):
        component_fn = component_from_script(path="./examples/ex11_using_scripts.py")

    @dsl.pipeline
    def pipeline_11():
        component_fn(["--hello", "kfp-backend"])

    _test_build_kfp_pipeline(pipeline_11())
